import importlib
import json
import sys
import threading
from unittest import mock

import pytest

import sentry_sdk
import sentry_sdk.integrations.quart as quart_sentry
from sentry_sdk import (
    capture_exception,
    capture_message,
    set_tag,
)
from sentry_sdk.integrations.logging import LoggingIntegration
from sentry_sdk.utils import SENSITIVE_DATA_SUBSTITUTE


def quart_app_factory():
    # These imports are inlined because the `test_quart_flask_patch` testcase
    # tests behavior that is triggered by importing a package before any Quart
    # imports happen, so we can't have these on the module level
    from quart import Quart

    try:
        from quart_auth import QuartAuth

        auth_manager = QuartAuth()
    except ImportError:
        from quart_auth import AuthManager

        auth_manager = AuthManager()

    app = Quart(__name__)
    app.debug = False
    app.config["TESTING"] = False
    app.secret_key = "haha"

    auth_manager.init_app(app)

    @app.route("/message")
    async def hi():
        capture_message("hi")
        return "ok"

    @app.route("/message/<message_id>")
    async def hi_with_id(message_id):
        capture_message("hi with id")
        return "ok with id"

    @app.get("/sync/thread_ids")
    def _thread_ids_sync():
        return {
            "main": str(threading.main_thread().ident),
            "active": str(threading.current_thread().ident),
        }

    @app.get("/async/thread_ids")
    async def _thread_ids_async():
        return {
            "main": str(threading.main_thread().ident),
            "active": str(threading.current_thread().ident),
        }

    return app


@pytest.fixture(params=("manual",))
def integration_enabled_params(request):
    if request.param == "manual":
        return {"integrations": [quart_sentry.QuartIntegration()]}
    else:
        raise ValueError(request.param)


@pytest.mark.asyncio
@pytest.mark.forked
@pytest.mark.skipif(
    not importlib.util.find_spec("quart_flask_patch"),
    reason="requires quart_flask_patch",
)
@pytest.mark.skipif(
    sys.version_info >= (3, 14),
    reason="quart_flask_patch not working on 3.14 (yet?)",
)
async def test_quart_flask_patch(sentry_init, capture_events, reset_integrations):
    # This testcase is forked because `import quart_flask_patch` needs to run
    # before anything else Quart-related is imported (since it monkeypatches
    # some things) and we don't want this to affect other testcases.
    #
    # It's also important this testcase be run before any other testcase
    # that uses `quart_app_factory`.
    import quart_flask_patch  # noqa: F401

    app = quart_app_factory()
    sentry_init(
        integrations=[quart_sentry.QuartIntegration()],
    )

    @app.route("/")
    async def index():
        1 / 0

    events = capture_events()

    client = app.test_client()
    try:
        await client.get("/")
    except ZeroDivisionError:
        pass

    (event,) = events
    assert event["exception"]["values"][0]["mechanism"]["type"] == "quart"


@pytest.mark.asyncio
async def test_has_context(sentry_init, capture_events):
    sentry_init(integrations=[quart_sentry.QuartIntegration()])
    app = quart_app_factory()
    events = capture_events()

    client = app.test_client()
    response = await client.get("/message")
    assert response.status_code == 200

    (event,) = events
    assert event["transaction"] == "hi"
    assert "data" not in event["request"]
    assert event["request"]["url"] == "http://localhost/message"


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "url,transaction_style,expected_transaction,expected_source",
    [
        ("/message", "endpoint", "hi", "component"),
        ("/message", "url", "/message", "route"),
        ("/message/123456", "endpoint", "hi_with_id", "component"),
        ("/message/123456", "url", "/message/<message_id>", "route"),
    ],
)
async def test_transaction_style(
    sentry_init,
    capture_events,
    url,
    transaction_style,
    expected_transaction,
    expected_source,
):
    sentry_init(
        integrations=[
            quart_sentry.QuartIntegration(transaction_style=transaction_style)
        ]
    )
    app = quart_app_factory()
    events = capture_events()

    client = app.test_client()
    response = await client.get(url)
    assert response.status_code == 200

    (event,) = events
    assert event["transaction"] == expected_transaction


@pytest.mark.asyncio
async def test_errors(
    sentry_init,
    capture_exceptions,
    capture_events,
    integration_enabled_params,
):
    sentry_init(**integration_enabled_params)
    app = quart_app_factory()

    @app.route("/")
    async def index():
        1 / 0

    exceptions = capture_exceptions()
    events = capture_events()

    client = app.test_client()
    try:
        await client.get("/")
    except ZeroDivisionError:
        pass

    (exc,) = exceptions
    assert isinstance(exc, ZeroDivisionError)

    (event,) = events
    assert event["exception"]["values"][0]["mechanism"]["type"] == "quart"


@pytest.mark.asyncio
async def test_quart_auth_not_installed(
    sentry_init, capture_events, monkeypatch, integration_enabled_params
):
    sentry_init(**integration_enabled_params)
    app = quart_app_factory()

    monkeypatch.setattr(quart_sentry, "quart_auth", None)

    events = capture_events()

    client = app.test_client()
    await client.get("/message")

    (event,) = events
    assert event.get("user", {}).get("id") is None


@pytest.mark.asyncio
async def test_quart_auth_not_configured(
    sentry_init, capture_events, monkeypatch, integration_enabled_params
):
    sentry_init(**integration_enabled_params)
    app = quart_app_factory()

    assert quart_sentry.quart_auth

    events = capture_events()
    client = app.test_client()
    await client.get("/message")

    (event,) = events
    assert event.get("user", {}).get("id") is None


@pytest.mark.asyncio
async def test_quart_auth_partially_configured(
    sentry_init, capture_events, monkeypatch, integration_enabled_params
):
    sentry_init(**integration_enabled_params)
    app = quart_app_factory()

    events = capture_events()

    client = app.test_client()
    await client.get("/message")

    (event,) = events
    assert event.get("user", {}).get("id") is None


@pytest.mark.asyncio
@pytest.mark.parametrize("send_default_pii", [True, False])
@pytest.mark.parametrize("user_id", [None, "42", "3"])
async def test_quart_auth_configured(
    send_default_pii,
    sentry_init,
    user_id,
    capture_events,
    monkeypatch,
    integration_enabled_params,
):
    from quart_auth import AuthUser, login_user

    sentry_init(send_default_pii=send_default_pii, **integration_enabled_params)
    app = quart_app_factory()

    @app.route("/login")
    async def login():
        if user_id is not None:
            login_user(AuthUser(user_id))
        return "ok"

    events = capture_events()

    client = app.test_client()
    assert (await client.get("/login")).status_code == 200
    assert not events

    assert (await client.get("/message")).status_code == 200

    (event,) = events
    if user_id is None or not send_default_pii:
        assert event.get("user", {}).get("id") is None
    else:
        assert event["user"]["id"] == str(user_id)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "integrations",
    [
        [quart_sentry.QuartIntegration()],
        [quart_sentry.QuartIntegration(), LoggingIntegration(event_level="ERROR")],
    ],
)
async def test_errors_not_reported_twice(sentry_init, integrations, capture_events):
    sentry_init(integrations=integrations)
    app = quart_app_factory()

    @app.route("/")
    async def index():
        try:
            1 / 0
        except Exception as e:
            app.logger.exception(e)
            raise e

    events = capture_events()

    client = app.test_client()
    # with pytest.raises(ZeroDivisionError):
    await client.get("/")

    assert len(events) == 1


@pytest.mark.asyncio
async def test_logging(sentry_init, capture_events):
    # ensure that Quart's logger magic doesn't break ours
    sentry_init(
        integrations=[
            quart_sentry.QuartIntegration(),
            LoggingIntegration(event_level="ERROR"),
        ]
    )
    app = quart_app_factory()

    @app.route("/")
    async def index():
        app.logger.error("hi")
        return "ok"

    events = capture_events()

    client = app.test_client()
    await client.get("/")

    (event,) = events
    assert event["level"] == "error"


@pytest.mark.asyncio
async def test_no_errors_without_request(sentry_init):
    sentry_init(integrations=[quart_sentry.QuartIntegration()])
    app = quart_app_factory()

    async with app.app_context():
        capture_exception(ValueError())


def test_cli_commands_raise():
    app = quart_app_factory()

    if not hasattr(app, "cli"):
        pytest.skip("Too old quart version")

    from quart.cli import ScriptInfo

    @app.cli.command()
    def foo():
        1 / 0

    with pytest.raises(ZeroDivisionError):
        app.cli.main(
            args=["foo"], prog_name="myapp", obj=ScriptInfo(create_app=lambda _: app)
        )


@pytest.mark.asyncio
async def test_500(sentry_init):
    sentry_init(integrations=[quart_sentry.QuartIntegration()])
    app = quart_app_factory()

    @app.route("/")
    async def index():
        1 / 0

    @app.errorhandler(500)
    async def error_handler(err):
        return "Sentry error."

    client = app.test_client()
    response = await client.get("/")

    assert (await response.get_data(as_text=True)) == "Sentry error."


@pytest.mark.asyncio
async def test_error_in_errorhandler(sentry_init, capture_events):
    sentry_init(integrations=[quart_sentry.QuartIntegration()])
    app = quart_app_factory()

    @app.route("/")
    async def index():
        raise ValueError()

    @app.errorhandler(500)
    async def error_handler(err):
        1 / 0

    events = capture_events()

    client = app.test_client()

    with pytest.raises(ZeroDivisionError):
        await client.get("/")

    event1, event2 = events

    (exception,) = event1["exception"]["values"]
    assert exception["type"] == "ValueError"

    exception = event2["exception"]["values"][-1]
    assert exception["type"] == "ZeroDivisionError"


@pytest.mark.asyncio
async def test_bad_request_not_captured(sentry_init, capture_events):
    from quart import abort

    sentry_init(integrations=[quart_sentry.QuartIntegration()])
    app = quart_app_factory()
    events = capture_events()

    @app.route("/")
    async def index():
        abort(400)

    client = app.test_client()

    await client.get("/")

    assert not events


@pytest.mark.asyncio
async def test_does_not_leak_scope(sentry_init, capture_events):
    from quart import Response, stream_with_context

    sentry_init(integrations=[quart_sentry.QuartIntegration()])
    app = quart_app_factory()
    events = capture_events()

    sentry_sdk.get_isolation_scope().set_tag("request_data", False)

    @app.route("/")
    async def index():
        sentry_sdk.get_isolation_scope().set_tag("request_data", True)

        async def generate():
            for row in range(1000):
                assert sentry_sdk.get_isolation_scope()._tags["request_data"]

                yield str(row) + "\n"

        return Response(stream_with_context(generate)(), mimetype="text/csv")

    client = app.test_client()
    response = await client.get("/")
    assert (await response.get_data(as_text=True)) == "".join(
        str(row) + "\n" for row in range(1000)
    )
    assert not events
    assert not sentry_sdk.get_isolation_scope()._tags["request_data"]


@pytest.mark.asyncio
async def test_scoped_test_client(sentry_init):
    sentry_init(integrations=[quart_sentry.QuartIntegration()])
    app = quart_app_factory()

    @app.route("/")
    async def index():
        return "ok"

    async with app.test_client() as client:
        response = await client.get("/")
        assert response.status_code == 200


@pytest.mark.asyncio
@pytest.mark.parametrize("exc_cls", [ZeroDivisionError, Exception])
async def test_errorhandler_for_exception_swallows_exception(
    sentry_init, capture_events, exc_cls
):
    # In contrast to error handlers for a status code, error
    # handlers for exceptions can swallow the exception (this is
    # just how the Quart signal works)
    sentry_init(integrations=[quart_sentry.QuartIntegration()])
    app = quart_app_factory()
    events = capture_events()

    @app.route("/")
    async def index():
        1 / 0

    @app.errorhandler(exc_cls)
    async def zerodivision(e):
        return "ok"

    async with app.test_client() as client:
        response = await client.get("/")
        assert response.status_code == 200

    assert not events


@pytest.mark.asyncio
async def test_tracing_success(sentry_init, capture_events):
    sentry_init(traces_sample_rate=1.0, integrations=[quart_sentry.QuartIntegration()])
    app = quart_app_factory()

    @app.before_request
    async def _():
        set_tag("before_request", "yes")

    @app.route("/message_tx")
    async def hi_tx():
        set_tag("view", "yes")
        capture_message("hi")
        return "ok"

    events = capture_events()

    async with app.test_client() as client:
        response = await client.get("/message_tx")
        assert response.status_code == 200

    message_event, transaction_event = events

    assert transaction_event["type"] == "transaction"
    assert transaction_event["transaction"] == "hi_tx"
    assert transaction_event["tags"]["view"] == "yes"
    assert transaction_event["tags"]["before_request"] == "yes"

    assert message_event["message"] == "hi"
    assert message_event["transaction"] == "hi_tx"
    assert message_event["tags"]["view"] == "yes"
    assert message_event["tags"]["before_request"] == "yes"


@pytest.mark.asyncio
async def test_tracing_error(sentry_init, capture_events):
    sentry_init(traces_sample_rate=1.0, integrations=[quart_sentry.QuartIntegration()])
    app = quart_app_factory()

    events = capture_events()

    @app.route("/error")
    async def error():
        1 / 0

    async with app.test_client() as client:
        response = await client.get("/error")
        assert response.status_code == 500

    error_event, transaction_event = events

    assert transaction_event["type"] == "transaction"
    assert transaction_event["transaction"] == "error"

    assert error_event["transaction"] == "error"
    (exception,) = error_event["exception"]["values"]
    assert exception["type"] == "ZeroDivisionError"


@pytest.mark.asyncio
async def test_class_based_views(sentry_init, capture_events):
    from quart.views import View

    sentry_init(integrations=[quart_sentry.QuartIntegration()])
    app = quart_app_factory()
    events = capture_events()

    @app.route("/")
    class HelloClass(View):
        methods = ["GET"]

        async def dispatch_request(self):
            capture_message("hi")
            return "ok"

    app.add_url_rule("/hello-class/", view_func=HelloClass.as_view("hello_class"))

    async with app.test_client() as client:
        response = await client.get("/hello-class/")
        assert response.status_code == 200

    (event,) = events

    assert event["message"] == "hi"
    assert event["transaction"] == "hello_class"


@pytest.mark.parametrize("endpoint", ["/sync/thread_ids", "/async/thread_ids"])
@pytest.mark.asyncio
async def test_active_thread_id(
    sentry_init, capture_envelopes, teardown_profiling, endpoint
):
    with mock.patch(
        "sentry_sdk.profiler.transaction_profiler.PROFILE_MINIMUM_SAMPLES", 0
    ):
        sentry_init(
            traces_sample_rate=1.0,
            profiles_sample_rate=1.0,
        )
        app = quart_app_factory()

        envelopes = capture_envelopes()

        async with app.test_client() as client:
            response = await client.get(endpoint)
            assert response.status_code == 200

        data = json.loads(await response.get_data(as_text=True))

        envelopes = [envelope for envelope in envelopes]
        assert len(envelopes) == 1

        profiles = [item for item in envelopes[0].items if item.type == "profile"]
        assert len(profiles) == 1, envelopes[0].items

        for item in profiles:
            transactions = item.payload.json["transactions"]
            assert len(transactions) == 1
            assert str(data["active"]) == transactions[0]["active_thread_id"]

        transactions = [
            item for item in envelopes[0].items if item.type == "transaction"
        ]
        assert len(transactions) == 1

        for item in transactions:
            transaction = item.payload.json
            trace_context = transaction["contexts"]["trace"]
            assert str(data["active"]) == trace_context["data"]["thread.id"]


@pytest.mark.parametrize("endpoint", ["/sync/thread_ids", "/async/thread_ids"])
@pytest.mark.asyncio
async def test_active_thread_id_span_streaming(
    sentry_init, capture_items, teardown_profiling, endpoint
):
    with mock.patch(
        "sentry_sdk.profiler.transaction_profiler.PROFILE_MINIMUM_SAMPLES", 0
    ):
        sentry_init(
            traces_sample_rate=1.0,
            profiles_sample_rate=1.0,
            trace_lifecycle="stream",
        )
        app = quart_app_factory()

        items = capture_items("span")

        async with app.test_client() as client:
            response = await client.get(endpoint)
            assert response.status_code == 200

        data = json.loads(await response.get_data(as_text=True))

        sentry_sdk.flush()

        spans = [item.payload for item in items]
        assert len(spans) == 1

        segment = spans[0]
        assert str(data["active"]) == segment["attributes"]["thread.id"]


@pytest.mark.asyncio
async def test_span_origin(sentry_init, capture_events):
    sentry_init(
        integrations=[quart_sentry.QuartIntegration()],
        traces_sample_rate=1.0,
    )
    app = quart_app_factory()
    events = capture_events()

    client = app.test_client()
    await client.get("/message")

    (_, event) = events

    assert event["contexts"]["trace"]["origin"] == "auto.http.quart"


@pytest.mark.asyncio
async def test_span_streaming_basic(sentry_init, capture_items):
    sentry_init(
        integrations=[quart_sentry.QuartIntegration()],
        traces_sample_rate=1.0,
        trace_lifecycle="stream",
    )
    items = capture_items("span")

    app = quart_app_factory()
    client = app.test_client()
    response = await client.get("/message")
    assert response.status_code == 200

    sentry_sdk.flush()

    spans = [item.payload for item in items]
    assert len(spans) == 1

    segment = spans[0]
    assert segment["is_segment"] is True
    assert "parent_span_id" not in segment
    assert segment["status"] == "ok"
    assert segment["attributes"]["sentry.op"] == "http.server"
    assert segment["attributes"]["sentry.origin"] == "auto.http.quart"
    assert segment["attributes"]["http.request.method"] == "GET"
    assert segment["name"] == "hi"


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "url,transaction_style,expected_name,expected_source",
    [
        ("/message", "endpoint", "hi", "component"),
        ("/message", "url", "/message", "route"),
        ("/message/123456", "endpoint", "hi_with_id", "component"),
        ("/message/123456", "url", "/message/<message_id>", "route"),
    ],
)
async def test_span_streaming_transaction_style(
    sentry_init,
    capture_items,
    url,
    transaction_style,
    expected_name,
    expected_source,
):
    sentry_init(
        integrations=[
            quart_sentry.QuartIntegration(transaction_style=transaction_style)
        ],
        traces_sample_rate=1.0,
        trace_lifecycle="stream",
    )
    items = capture_items("span")

    app = quart_app_factory()
    client = app.test_client()
    response = await client.get(url)
    assert response.status_code == 200

    sentry_sdk.flush()

    spans = [item.payload for item in items]
    assert len(spans) == 1

    segment = spans[0]
    assert segment["is_segment"] is True
    assert segment["name"] == expected_name
    assert segment["attributes"]["sentry.segment.name.source"] == expected_source


@pytest.mark.asyncio
async def test_span_streaming_with_error(sentry_init, capture_items):
    sentry_init(
        integrations=[quart_sentry.QuartIntegration()],
        traces_sample_rate=1.0,
        trace_lifecycle="stream",
    )
    items = capture_items("event", "span")

    app = quart_app_factory()

    @app.route("/error")
    async def error():
        1 / 0

    client = app.test_client()
    try:
        await client.get("/error")
    except ZeroDivisionError:
        pass

    sentry_sdk.flush()

    events = [item.payload for item in items if item.type == "event"]
    spans = [item.payload for item in items if item.type == "span"]

    assert len(events) == 1
    assert len(spans) == 1

    error_event = events[0]
    segment = spans[0]

    assert segment["trace_id"] == error_event["contexts"]["trace"]["trace_id"]
    assert segment["is_segment"] is True
    assert segment["status"] == "error"

    assert "parent_span_id" not in segment

    assert error_event["contexts"]["trace"]["span_id"] == segment["span_id"]
    assert error_event["exception"]["values"][0]["mechanism"]["type"] == "quart"
    assert error_event["exception"]["values"][0]["mechanism"]["handled"] is False


@pytest.mark.asyncio
async def test_span_streaming_request_attributes_no_pii(sentry_init, capture_items):
    sentry_init(
        integrations=[quart_sentry.QuartIntegration()],
        traces_sample_rate=1.0,
        send_default_pii=False,
        trace_lifecycle="stream",
    )
    items = capture_items("span")

    app = quart_app_factory()
    client = app.test_client()
    response = await client.get("/message?foo=bar")
    assert response.status_code == 200

    sentry_sdk.flush()

    spans = [item.payload for item in items]
    assert len(spans) == 1

    segment = spans[0]
    assert segment["attributes"]["http.request.method"] == "GET"
    assert "http.request.header.host" in segment["attributes"]

    assert "url.full" not in segment["attributes"]
    assert "url.path" not in segment["attributes"]
    assert "url.query" not in segment["attributes"]
    assert "client.address" not in segment["attributes"]
    assert "user.ip_address" not in segment["attributes"]


@pytest.mark.asyncio
async def test_span_streaming_request_attributes_with_pii(sentry_init, capture_items):
    sentry_init(
        integrations=[quart_sentry.QuartIntegration()],
        traces_sample_rate=1.0,
        send_default_pii=True,
        trace_lifecycle="stream",
    )
    items = capture_items("span")

    app = quart_app_factory()
    client = app.test_client()
    response = await client.get("/message?foo=bar&baz=qux")
    assert response.status_code == 200

    sentry_sdk.flush()

    spans = [item.payload for item in items]
    assert len(spans) == 1

    segment = spans[0]
    assert segment["attributes"]["http.request.method"] == "GET"
    assert "http.request.header.host" in segment["attributes"]

    assert (
        segment["attributes"]["url.full"] == "http://localhost/message?foo=bar&baz=qux"
    )
    assert segment["attributes"]["url.path"] == "/message"
    assert segment["attributes"]["url.query"] == "foo=bar&baz=qux"
    assert "client.address" in segment["attributes"]
    assert "user.ip_address" in segment["attributes"]


@pytest.mark.asyncio
async def test_span_streaming_sensitive_header_scrubbing(sentry_init, capture_items):
    sentry_init(
        integrations=[quart_sentry.QuartIntegration()],
        traces_sample_rate=1.0,
        send_default_pii=False,
        trace_lifecycle="stream",
    )
    items = capture_items("span")

    app = quart_app_factory()
    client = app.test_client()
    response = await client.get(
        "/message",
        headers={
            "Authorization": "Bearer secret-token",
            "X-Custom-Header": "passthrough",
        },
    )
    assert response.status_code == 200

    sentry_sdk.flush()

    spans = [item.payload for item in items]
    assert len(spans) == 1

    segment = spans[0]
    assert (
        segment["attributes"]["http.request.header.authorization"]
        == SENSITIVE_DATA_SUBSTITUTE
    )
    assert segment["attributes"]["http.request.header.x-custom-header"] == "passthrough"


@pytest.mark.asyncio
@pytest.mark.parametrize("send_default_pii", [True, False])
@pytest.mark.parametrize("user_id", [None, "42"])
async def test_span_streaming_quart_auth_user_id(
    send_default_pii,
    sentry_init,
    user_id,
    capture_items,
):
    from quart_auth import AuthUser, login_user

    sentry_init(
        integrations=[quart_sentry.QuartIntegration()],
        traces_sample_rate=1.0,
        send_default_pii=send_default_pii,
        trace_lifecycle="stream",
    )
    items = capture_items("span")

    app = quart_app_factory()

    @app.route("/login")
    async def login():
        if user_id is not None:
            login_user(AuthUser(user_id))
        return "ok"

    client = app.test_client()
    assert (await client.get("/login")).status_code == 200
    assert (await client.get("/message")).status_code == 200

    sentry_sdk.flush()

    spans = [item.payload for item in items]
    assert len(spans) == 2

    segment = spans[1]
    if send_default_pii and user_id is not None:
        assert segment["attributes"]["user.id"] == user_id
    else:
        assert "user.id" not in segment.get("attributes", {})


@pytest.mark.asyncio
async def test_span_streaming_sensitive_header_passthrough_with_pii(
    sentry_init, capture_items
):
    sentry_init(
        integrations=[quart_sentry.QuartIntegration()],
        traces_sample_rate=1.0,
        send_default_pii=True,
        trace_lifecycle="stream",
    )
    items = capture_items("span")

    app = quart_app_factory()
    client = app.test_client()
    response = await client.get(
        "/message",
        headers={"Authorization": "Bearer secret-token"},
    )
    assert response.status_code == 200

    sentry_sdk.flush()

    spans = [item.payload for item in items]
    assert len(spans) == 1

    segment = spans[0]
    assert (
        segment["attributes"]["http.request.header.authorization"]
        == "Bearer secret-token"
    )



GEN_AI_SYSTEM = "gcp.gemini"

# Mapping of tool attributes to their descriptions
# These are all tools that are available in the Google GenAI API
TOOL_ATTRIBUTES_MAP = {
    "google_search_retrieval": "Google Search retrieval tool",
    "google_search": "Google Search tool",
    "retrieval": "Retrieval tool",
    "enterprise_web_search": "Enterprise web search tool",
    "google_maps": "Google Maps tool",
    "code_execution": "Code execution tool",
    "computer_use": "Computer use tool",
}

IDENTIFIER = "google_genai"
ORIGIN = f"auto.ai.{IDENTIFIER}"



import logging
import sys

from sentry_sdk import capture_event, capture_exception, start_span, start_transaction
from sentry_sdk.scrubber import EventScrubber
from sentry_sdk.utils import event_from_exception
from tests.conftest import ApproxDict

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


def test_request_scrubbing(sentry_init, capture_events):
    sentry_init()
    events = capture_events()

    try:
        1 / 0
    except ZeroDivisionError:
        ev, _hint = event_from_exception(sys.exc_info())

        ev["request"] = {
            "headers": {
                "COOKIE": "secret",
                "authorization": "Bearer bla",
                "ORIGIN": "google.com",
                "ip_address": "127.0.0.1",
            },
            "cookies": {
                "sessionid": "secret",
                "foo": "bar",
            },
            "data": {
                "token": "secret",
                "foo": "bar",
            },
        }

        capture_event(ev)

    (event,) = events

    assert event["request"] == {
        "headers": {
            "COOKIE": "[Filtered]",
            "authorization": "[Filtered]",
            "ORIGIN": "google.com",
            "ip_address": "[Filtered]",
        },
        "cookies": {"sessionid": "[Filtered]", "foo": "bar"},
        "data": {"token": "[Filtered]", "foo": "bar"},
    }

    assert event["_meta"]["request"] == {
        "headers": {
            "COOKIE": {"": {"rem": [["!config", "s"]]}},
            "authorization": {"": {"rem": [["!config", "s"]]}},
            "ip_address": {"": {"rem": [["!config", "s"]]}},
        },
        "cookies": {"sessionid": {"": {"rem": [["!config", "s"]]}}},
        "data": {"token": {"": {"rem": [["!config", "s"]]}}},
    }


def test_ip_address_not_scrubbed_when_pii_enabled(sentry_init, capture_events):
    sentry_init(send_default_pii=True)
    events = capture_events()

    try:
        1 / 0
    except ZeroDivisionError:
        ev, _hint = event_from_exception(sys.exc_info())

        ev["request"] = {"headers": {"COOKIE": "secret", "ip_address": "127.0.0.1"}}

        capture_event(ev)

    (event,) = events

    assert event["request"] == {
        "headers": {"COOKIE": "[Filtered]", "ip_address": "127.0.0.1"}
    }

    assert event["_meta"]["request"] == {
        "headers": {
            "COOKIE": {"": {"rem": [["!config", "s"]]}},
        }
    }


def test_user_ip_address_removed_when_pii_disabled(sentry_init, capture_events):
    sentry_init()
    events = capture_events()

    try:
        1 / 0
    except ZeroDivisionError:
        ev, _hint = event_from_exception(sys.exc_info())
        ev["user"] = {"id": "42", "ip_address": "127.0.0.1"}

        capture_event(ev)

    (event,) = events

    assert event["user"] == {"id": "42"}
    assert "user" not in event.get("_meta", {})


def test_user_ip_address_not_removed_when_pii_enabled(sentry_init, capture_events):
    sentry_init(send_default_pii=True)
    events = capture_events()

    try:
        1 / 0
    except ZeroDivisionError:
        ev, _hint = event_from_exception(sys.exc_info())
        ev["user"] = {"id": "42", "ip_address": "127.0.0.1"}

        capture_event(ev)

    (event,) = events

    assert event["user"] == {"id": "42", "ip_address": "127.0.0.1"}
    assert "user" not in event.get("_meta", {})


def test_stack_var_scrubbing(sentry_init, capture_events):
    sentry_init()
    events = capture_events()

    try:
        password = "supersecret"  # noqa
        api_key = "1231231231"  # noqa
        safe = "keepthis"  # noqa
        1 / 0
    except ZeroDivisionError:
        capture_exception()

    (event,) = events

    frames = event["exception"]["values"][0]["stacktrace"]["frames"]
    (frame,) = frames
    assert frame["vars"]["password"] == "[Filtered]"
    assert frame["vars"]["api_key"] == "[Filtered]"
    assert frame["vars"]["safe"] == "'keepthis'"

    meta = event["_meta"]["exception"]["values"]["0"]["stacktrace"]["frames"]["0"][
        "vars"
    ]
    assert meta == {
        "password": {"": {"rem": [["!config", "s"]]}},
        "api_key": {"": {"rem": [["!config", "s"]]}},
    }


def test_breadcrumb_extra_scrubbing(sentry_init, capture_events):
    sentry_init(max_breadcrumbs=2)
    events = capture_events()
    logger.info("breadcrumb 1", extra=dict(foo=1, password="secret"))
    logger.info("breadcrumb 2", extra=dict(bar=2, auth="secret"))
    logger.info("breadcrumb 3", extra=dict(foobar=3, password="secret"))
    logger.critical("whoops", extra=dict(bar=69, auth="secret"))

    (event,) = events

    assert event["extra"]["bar"] == 69
    assert event["extra"]["auth"] == "[Filtered]"
    assert event["breadcrumbs"]["values"][0]["data"] == {
        "bar": 2,
        "auth": "[Filtered]",
    }
    assert event["breadcrumbs"]["values"][1]["data"] == {
        "foobar": 3,
        "password": "[Filtered]",
    }

    assert event["_meta"]["extra"]["auth"] == {"": {"rem": [["!config", "s"]]}}
    assert event["_meta"]["breadcrumbs"] == {
        "": {"len": 3},
        "values": {
            "0": {"data": {"auth": {"": {"rem": [["!config", "s"]]}}}},
            "1": {"data": {"password": {"": {"rem": [["!config", "s"]]}}}},
        },
    }


def test_span_data_scrubbing(sentry_init, capture_events):
    sentry_init(traces_sample_rate=1.0)
    events = capture_events()

    with start_transaction(name="hi"):
        with start_span(op="foo", name="bar") as span:
            span.set_data("password", "secret")
            span.set_data("datafoo", "databar")

    (event,) = events
    assert event["spans"][0]["data"] == ApproxDict(
        {"password": "[Filtered]", "datafoo": "databar"}
    )
    assert event["_meta"]["spans"] == {
        "0": {"data": {"password": {"": {"rem": [["!config", "s"]]}}}}
    }


def test_custom_denylist(sentry_init, capture_events):
    sentry_init(
        event_scrubber=EventScrubber(
            denylist=["my_sensitive_var"], pii_denylist=["my_pii_var"]
        )
    )
    events = capture_events()

    try:
        my_sensitive_var = "secret"  # noqa
        my_pii_var = "jane.doe"  # noqa
        safe = "keepthis"  # noqa
        1 / 0
    except ZeroDivisionError:
        capture_exception()

    (event,) = events

    frames = event["exception"]["values"][0]["stacktrace"]["frames"]
    (frame,) = frames
    assert frame["vars"]["my_sensitive_var"] == "[Filtered]"
    assert frame["vars"]["my_pii_var"] == "[Filtered]"
    assert frame["vars"]["safe"] == "'keepthis'"

    meta = event["_meta"]["exception"]["values"]["0"]["stacktrace"]["frames"]["0"][
        "vars"
    ]
    assert meta == {
        "my_sensitive_var": {"": {"rem": [["!config", "s"]]}},
        "my_pii_var": {"": {"rem": [["!config", "s"]]}},
    }


def test_scrubbing_doesnt_affect_local_vars(sentry_init, capture_events):
    sentry_init()
    events = capture_events()

    try:
        password = "cat123"
        1 / 0
    except ZeroDivisionError:
        capture_exception()

    (event,) = events

    frames = event["exception"]["values"][0]["stacktrace"]["frames"]
    (frame,) = frames
    assert frame["vars"]["password"] == "[Filtered]"
    assert password == "cat123"


def test_recursive_event_scrubber(sentry_init, capture_events):
    sentry_init(event_scrubber=EventScrubber(recursive=True))
    events = capture_events()
    complex_structure = {
        "deep": {
            "deeper": [{"deepest": {"password": "my_darkest_secret"}}],
        },
    }

    capture_event({"extra": complex_structure})

    (event,) = events
    assert event["extra"]["deep"]["deeper"][0]["deepest"]["password"] == "'[Filtered]'"


def test_recursive_scrubber_does_not_override_original(sentry_init, capture_events):
    sentry_init(event_scrubber=EventScrubber(recursive=True))
    events = capture_events()

    data = {"csrf": "secret"}
    try:
        raise RuntimeError("An error")
    except Exception:
        capture_exception()

    (event,) = events
    frames = event["exception"]["values"][0]["stacktrace"]["frames"]
    (frame,) = frames
    assert data["csrf"] == "secret"
    assert frame["vars"]["data"]["csrf"] == "[Filtered]"



import datetime
import os
from contextlib import contextmanager
from http.server import BaseHTTPRequestHandler, HTTPServer
from threading import Thread
from unittest import mock

import pytest
from pyreqwest.client import ClientBuilder, SyncClientBuilder
from pyreqwest.simple.request import pyreqwest_get as async_pyreqwest_get
from pyreqwest.simple.sync_request import pyreqwest_get as sync_pyreqwest_get

import sentry_sdk
from sentry_sdk import start_transaction
from sentry_sdk.consts import MATCH_ALL, SPANDATA
from sentry_sdk.integrations.pyreqwest import PyreqwestIntegration
from tests.conftest import get_free_port


class PyreqwestMockHandler(BaseHTTPRequestHandler):
    captured_requests = []

    def do_GET(self) -> None:
        self.captured_requests.append(
            {
                "path": self.path,
                "headers": {k.lower(): v for k, v in self.headers.items()},
            }
        )

        code = 200
        if "/status/" in self.path:
            try:
                code = int(self.path.split("/")[-1])
            except (ValueError, IndexError):
                code = 200

        self.send_response(code)
        self.end_headers()
        self.wfile.write(b"OK")

    def log_message(self, format: str, *args: object) -> None:
        pass


@pytest.fixture(scope="module")
def server_port():
    port = get_free_port()
    server = HTTPServer(("localhost", port), PyreqwestMockHandler)
    thread = Thread(target=server.serve_forever)
    thread.daemon = True
    thread.start()
    yield port
    server.shutdown()


@pytest.fixture(autouse=True)
def clear_captured_requests():
    PyreqwestMockHandler.captured_requests.clear()


@pytest.mark.parametrize("send_default_pii", [True, False])
@pytest.mark.parametrize("span_streaming", [True, False])
def test_sync_client_spans(
    sentry_init,
    capture_events,
    capture_items,
    server_port,
    span_streaming,
    send_default_pii,
):
    sentry_init(
        integrations=[PyreqwestIntegration()],
        traces_sample_rate=1.0,
        send_default_pii=send_default_pii,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    url = f"http://localhost:{server_port}/hello?q=test#frag"
    if span_streaming:
        items = capture_items("span")

        with sentry_sdk.traces.start_span(name="custom parent"):
            client = SyncClientBuilder().build()
            response = client.get(url).build().send()
            assert response.status == 200

        sentry_sdk.flush()
        spans = [item.payload for item in items]
        assert len(spans) == 2
        span = spans[0]
        assert span["attributes"]["sentry.op"] == "http.client"
        assert span["name"] == f"GET http://localhost:{server_port}/hello"
        assert span["attributes"][SPANDATA.HTTP_REQUEST_METHOD] == "GET"
        assert span["attributes"][SPANDATA.HTTP_STATUS_CODE] == 200
        assert span["attributes"]["sentry.origin"] == "auto.http.pyreqwest"

        if send_default_pii:
            assert (
                span["attributes"]["url.full"]
                == f"http://localhost:{server_port}/hello"
            )
            assert span["attributes"][SPANDATA.URL_QUERY] == "q=test"
            assert span["attributes"][SPANDATA.URL_FRAGMENT] == "frag"
        else:
            assert "url.full" not in span["attributes"]
            assert SPANDATA.URL_QUERY not in span["attributes"]
            assert SPANDATA.URL_FRAGMENT not in span["attributes"]
    else:
        events = capture_events()

        with start_transaction(name="test_transaction"):
            client = SyncClientBuilder().build()
            response = client.get(url).build().send()
            assert response.status == 200

        (event,) = events
        assert len(event["spans"]) == 1
        span = event["spans"][0]
        assert span["op"] == "http.client"
        assert span["description"] == f"GET http://localhost:{server_port}/hello"
        assert span["data"]["url"] == f"http://localhost:{server_port}/hello"
        assert span["data"][SPANDATA.HTTP_METHOD] == "GET"
        assert span["data"][SPANDATA.HTTP_STATUS_CODE] == 200
        assert span["data"][SPANDATA.HTTP_QUERY] == "q=test"
        assert span["data"][SPANDATA.HTTP_FRAGMENT] == "frag"
        assert span["origin"] == "auto.http.pyreqwest"


@pytest.mark.asyncio
@pytest.mark.parametrize("send_default_pii", [True, False])
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_async_client_spans(
    sentry_init,
    capture_events,
    capture_items,
    server_port,
    span_streaming,
    send_default_pii,
):
    sentry_init(
        integrations=[PyreqwestIntegration()],
        traces_sample_rate=1.0,
        send_default_pii=send_default_pii,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    url = f"http://localhost:{server_port}/hello"
    if span_streaming:
        items = capture_items("span")

        async with ClientBuilder().build() as client:
            with sentry_sdk.traces.start_span(name="custom parent"):
                response = await client.get(url).build().send()
                assert response.status == 200

        sentry_sdk.flush()
        spans = [item.payload for item in items]
        assert len(spans) == 2
        span = spans[0]
        assert span["attributes"]["sentry.op"] == "http.client"
        assert span["name"] == f"GET {url}"
        assert span["attributes"][SPANDATA.HTTP_REQUEST_METHOD] == "GET"
        assert span["attributes"][SPANDATA.HTTP_STATUS_CODE] == 200
        assert span["attributes"]["sentry.origin"] == "auto.http.pyreqwest"

        if send_default_pii:
            assert span["attributes"]["url.full"] == url
        else:
            assert "url.full" not in span["attributes"]
    else:
        events = capture_events()

        async with ClientBuilder().build() as client:
            with start_transaction(name="test_transaction"):
                response = await client.get(url).build().send()
                assert response.status == 200

        (event,) = events
        assert len(event["spans"]) == 1
        span = event["spans"][0]
        assert span["op"] == "http.client"
        assert span["description"] == f"GET {url}"
        assert span["data"]["url"] == url
        assert span["data"][SPANDATA.HTTP_METHOD] == "GET"
        assert span["data"][SPANDATA.HTTP_STATUS_CODE] == 200
        assert span["origin"] == "auto.http.pyreqwest"


@pytest.mark.parametrize("send_default_pii", [True, False])
@pytest.mark.parametrize("span_streaming", [True, False])
def test_sync_simple_request_spans(
    sentry_init,
    capture_events,
    capture_items,
    server_port,
    span_streaming,
    send_default_pii,
):
    sentry_init(
        integrations=[PyreqwestIntegration()],
        traces_sample_rate=1.0,
        send_default_pii=send_default_pii,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    url = f"http://localhost:{server_port}/hello-simple"
    if span_streaming:
        items = capture_items("span")

        with sentry_sdk.traces.start_span(name="custom parent"):
            response = sync_pyreqwest_get(url).send()
            assert response.status == 200

        sentry_sdk.flush()
        spans = [item.payload for item in items]
        assert len(spans) == 2
        span = spans[0]
        assert span["attributes"]["sentry.op"] == "http.client"
        assert span["name"] == f"GET {url}"
        assert span["attributes"][SPANDATA.HTTP_REQUEST_METHOD] == "GET"
        assert span["attributes"][SPANDATA.HTTP_STATUS_CODE] == 200
        assert span["attributes"]["sentry.origin"] == "auto.http.pyreqwest"

        if send_default_pii:
            assert span["attributes"]["url.full"] == url
        else:
            assert "url.full" not in span["attributes"]
    else:
        events = capture_events()

        with start_transaction(name="test_transaction"):
            response = sync_pyreqwest_get(url).send()
            assert response.status == 200

        (event,) = events
        assert len(event["spans"]) == 1
        span = event["spans"][0]
        assert span["op"] == "http.client"
        assert span["description"] == f"GET {url}"
        assert span["data"]["url"] == url
        assert span["data"][SPANDATA.HTTP_METHOD] == "GET"
        assert span["data"][SPANDATA.HTTP_STATUS_CODE] == 200
        assert span["origin"] == "auto.http.pyreqwest"


@pytest.mark.asyncio
@pytest.mark.parametrize("send_default_pii", [True, False])
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_async_simple_request_spans(
    sentry_init,
    capture_events,
    capture_items,
    server_port,
    span_streaming,
    send_default_pii,
):
    sentry_init(
        integrations=[PyreqwestIntegration()],
        traces_sample_rate=1.0,
        send_default_pii=send_default_pii,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    url = f"http://localhost:{server_port}/hello-simple-async"
    if span_streaming:
        items = capture_items("span")

        with sentry_sdk.traces.start_span(name="custom parent"):
            response = await async_pyreqwest_get(url).send()
            assert response.status == 200

        sentry_sdk.flush()
        spans = [item.payload for item in items]
        assert len(spans) == 2
        span = spans[0]
        assert span["attributes"]["sentry.op"] == "http.client"
        assert span["name"] == f"GET {url}"
        assert span["attributes"][SPANDATA.HTTP_REQUEST_METHOD] == "GET"
        assert span["attributes"][SPANDATA.HTTP_STATUS_CODE] == 200
        assert span["attributes"]["sentry.origin"] == "auto.http.pyreqwest"

        if send_default_pii:
            assert span["attributes"]["url.full"] == url
        else:
            assert "url.full" not in span["attributes"]
    else:
        events = capture_events()

        with start_transaction(name="test_transaction"):
            response = await async_pyreqwest_get(url).send()
            assert response.status == 200

        (event,) = events
        assert len(event["spans"]) == 1
        span = event["spans"][0]
        assert span["op"] == "http.client"
        assert span["description"] == f"GET {url}"
        assert span["data"]["url"] == url
        assert span["data"][SPANDATA.HTTP_METHOD] == "GET"
        assert span["data"][SPANDATA.HTTP_STATUS_CODE] == 200
        assert span["origin"] == "auto.http.pyreqwest"


@pytest.mark.parametrize("span_streaming", [True, False])
def test_span_origin(
    sentry_init,
    capture_events,
    capture_items,
    server_port,
    span_streaming,
):
    sentry_init(
        integrations=[PyreqwestIntegration()],
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    url = f"http://localhost:{server_port}/origin"
    if span_streaming:
        items = capture_items("span")

        with sentry_sdk.traces.start_span(name="custom parent"):
            client = SyncClientBuilder().build()
            client.get(url).build().send()

        sentry_sdk.flush()
        spans = [item.payload for item in items]
        assert spans[0]["attributes"]["sentry.origin"] == "auto.http.pyreqwest"
    else:
        events = capture_events()

        with start_transaction(name="test_transaction"):
            client = SyncClientBuilder().build()
            client.get(url).build().send()

        (event,) = events
        assert event["spans"][0]["origin"] == "auto.http.pyreqwest"


@pytest.mark.parametrize("span_streaming", [True, False])
def test_outgoing_trace_headers(
    sentry_init,
    server_port,
    capture_items,
    span_streaming,
):
    sentry_init(
        integrations=[PyreqwestIntegration()],
        traces_sample_rate=1.0,
        trace_propagation_targets=["localhost"],
        trace_lifecycle="stream" if span_streaming else "static",
    )

    url = f"http://localhost:{server_port}/trace"
    if span_streaming:
        items = capture_items("span")

        with sentry_sdk.traces.start_span(
            name="custom parent",
        ):
            client = SyncClientBuilder().build()
            response = client.get(url).build().send()
            assert response.status == 200

        assert len(PyreqwestMockHandler.captured_requests) == 1
        headers = PyreqwestMockHandler.captured_requests[0]["headers"]

        sentry_sdk.flush()
        spans = [item.payload for item in items]
        http_span = next(
            span
            for span in spans
            if span["attributes"].get("sentry.op") == "http.client"
        )

        assert "sentry-trace" in headers
        assert headers["sentry-trace"].startswith(http_span["trace_id"])
        assert "baggage" in headers
        assert f"sentry-trace_id={http_span['trace_id']}" in headers["baggage"]
    else:
        with start_transaction(
            name="test_transaction", trace_id="01234567890123456789012345678901"
        ):
            client = SyncClientBuilder().build()
            response = client.get(url).build().send()
            assert response.status == 200

        assert len(PyreqwestMockHandler.captured_requests) == 1
        headers = PyreqwestMockHandler.captured_requests[0]["headers"]

        assert "sentry-trace" in headers
        assert headers["sentry-trace"].startswith("01234567890123456789012345678901")
        assert "baggage" in headers
        assert "sentry-trace_id=01234567890123456789012345678901" in headers["baggage"]


@pytest.mark.parametrize("span_streaming", [True, False])
def test_outgoing_trace_headers_append_to_baggage(
    sentry_init,
    server_port,
    capture_items,
    span_streaming,
):
    sentry_init(
        integrations=[PyreqwestIntegration()],
        traces_sample_rate=1.0,
        trace_propagation_targets=["localhost"],
        release="d08ebdb9309e1b004c6f52202de58a09c2268e42",
        trace_lifecycle="stream" if span_streaming else "static",
    )

    url = f"http://localhost:{server_port}/baggage"
    if span_streaming:
        items = capture_items("span")

        with mock.patch(
            "sentry_sdk.tracing_utils.Random.randrange", return_value=500000
        ):
            with sentry_sdk.traces.start_span(
                name="/interactions/other-dogs/new-dog",
                attributes={
                    "sentry.op": "greeting.sniff",
                },
            ):
                client = SyncClientBuilder().build()
                client.get(url).header("baggage", "custom=data").build().send()

        assert len(PyreqwestMockHandler.captured_requests) == 1
        headers = PyreqwestMockHandler.captured_requests[0]["headers"]

        sentry_sdk.flush()
        spans = [item.payload for item in items]
        http_span = next(
            span
            for span in spans
            if span["attributes"].get("sentry.op") == "http.client"
        )

        assert "baggage" in headers
        baggage = headers["baggage"]
        assert "custom=data" in baggage
        assert f"sentry-trace_id={http_span['trace_id']}" in baggage
    else:
        with mock.patch(
            "sentry_sdk.tracing_utils.Random.randrange", return_value=500000
        ):
            with start_transaction(
                name="/interactions/other-dogs/new-dog",
                op="greeting.sniff",
                trace_id="01234567890123456789012345678901",
            ):
                client = SyncClientBuilder().build()
                client.get(url).header("baggage", "custom=data").build().send()

        assert len(PyreqwestMockHandler.captured_requests) == 1
        headers = PyreqwestMockHandler.captured_requests[0]["headers"]

        assert "baggage" in headers
        baggage = headers["baggage"]
        assert "custom=data" in baggage
        assert "sentry-trace_id=01234567890123456789012345678901" in baggage
    assert "sentry-sample_rand=0.500000" in baggage
    assert "sentry-environment=production" in baggage
    assert "sentry-release=d08ebdb9309e1b004c6f52202de58a09c2268e42" in baggage
    assert "sentry-transaction=/interactions/other-dogs/new-dog" in baggage
    assert "sentry-sample_rate=1.0" in baggage
    assert "sentry-sampled=true" in baggage


@pytest.mark.parametrize(
    "trace_propagation_targets,trace_propagated",
    [
        [None, False],
        [[], False],
        [[MATCH_ALL], True],
        [["localhost"], True],
        [[r"https?:\/\/[\w\-]+(\.[\w\-]+)+\.net"], False],
    ],
)
@pytest.mark.parametrize("span_streaming", [True, False])
def test_trace_propagation_targets(
    sentry_init,
    server_port,
    trace_propagation_targets,
    trace_propagated,
    span_streaming,
):
    sentry_init(
        integrations=[PyreqwestIntegration()],
        trace_propagation_targets=trace_propagation_targets,
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    url = f"http://localhost:{server_port}/propagation"

    with start_transaction():
        client = SyncClientBuilder().build()
        client.get(url).build().send()

    assert len(PyreqwestMockHandler.captured_requests) == 1
    headers = PyreqwestMockHandler.captured_requests[0]["headers"]

    if trace_propagated:
        assert "sentry-trace" in headers
    else:
        assert "sentry-trace" not in headers


@pytest.mark.tests_internal_exceptions
@pytest.mark.parametrize("span_streaming", [True, False])
def test_omit_url_data_if_parsing_fails(
    sentry_init,
    capture_events,
    capture_items,
    server_port,
    span_streaming,
):
    sentry_init(
        integrations=[PyreqwestIntegration()],
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    url = f"http://localhost:{server_port}/parse-fail"
    if span_streaming:
        items = capture_items("span")

        with sentry_sdk.traces.start_span(name="custom parent"):
            with mock.patch(
                "sentry_sdk.integrations.pyreqwest.parse_url",
                side_effect=ValueError,
            ):
                client = SyncClientBuilder().build()
                client.get(url).build().send()

        sentry_sdk.flush()
        spans = [item.payload for item in items]
        span = spans[0]

        assert span["name"] == "GET [Filtered]"
        assert span["attributes"][SPANDATA.HTTP_REQUEST_METHOD] == "GET"
        assert span["attributes"][SPANDATA.HTTP_STATUS_CODE] == 200
        assert "url.full" not in span["attributes"]
        assert SPANDATA.URL_QUERY not in span["attributes"]
        assert SPANDATA.URL_FRAGMENT not in span["attributes"]
    else:
        events = capture_events()

        with start_transaction(name="test_transaction"):
            with mock.patch(
                "sentry_sdk.integrations.pyreqwest.parse_url",
                side_effect=ValueError,
            ):
                client = SyncClientBuilder().build()
                client.get(url).build().send()

        (event,) = events
        span = event["spans"][0]

        assert span["description"] == "GET [Filtered]"
        assert span["data"][SPANDATA.HTTP_METHOD] == "GET"
        assert span["data"][SPANDATA.HTTP_STATUS_CODE] == 200
        assert "url" not in span["data"]
        assert SPANDATA.HTTP_QUERY not in span["data"]
        assert SPANDATA.HTTP_FRAGMENT not in span["data"]


@pytest.mark.parametrize("span_streaming", [True, False])
def test_request_source_disabled(
    sentry_init,
    capture_events,
    capture_items,
    server_port,
    span_streaming,
):
    sentry_init(
        integrations=[PyreqwestIntegration()],
        traces_sample_rate=1.0,
        enable_http_request_source=False,
        http_request_source_threshold_ms=0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    url = f"http://localhost:{server_port}/hello"
    if span_streaming:
        items = capture_items("span")

        with sentry_sdk.traces.start_span(name="custom parent"):
            client = SyncClientBuilder().build()
            client.get(url).build().send()

        sentry_sdk.flush()
        spans = [item.payload for item in items]
        span = spans[0]
        data = span.get("attributes", {})

        assert SPANDATA.CODE_LINE_NUMBER not in data
        assert SPANDATA.CODE_NAMESPACE not in data
        assert SPANDATA.CODE_FILE_PATH not in data
    else:
        events = capture_events()

        with start_transaction(name="test_transaction"):
            client = SyncClientBuilder().build()
            client.get(url).build().send()

        (event,) = events
        span = event["spans"][0]
        data = span.get("data", {})

        assert SPANDATA.CODE_LINENO not in data
        assert SPANDATA.CODE_NAMESPACE not in data
        assert SPANDATA.CODE_FILEPATH not in data
    assert SPANDATA.CODE_FUNCTION not in data


@pytest.mark.parametrize("enable_http_request_source", [None, True])
@pytest.mark.parametrize("span_streaming", [True, False])
def test_request_source_enabled(
    sentry_init,
    capture_events,
    capture_items,
    server_port,
    enable_http_request_source,
    span_streaming,
):
    sentry_options = {
        "integrations": [PyreqwestIntegration()],
        "traces_sample_rate": 1.0,
        "http_request_source_threshold_ms": 0,
        "trace_lifecycle": "stream" if span_streaming else "static",
    }
    if enable_http_request_source is not None:
        sentry_options["enable_http_request_source"] = enable_http_request_source

    sentry_init(**sentry_options)

    url = f"http://localhost:{server_port}/hello"
    if span_streaming:
        items = capture_items("span")

        with sentry_sdk.traces.start_span(name="custom parent"):
            client = SyncClientBuilder().build()
            client.get(url).build().send()

        sentry_sdk.flush()
        spans = [item.payload for item in items]
        span = spans[0]
        data = span.get("attributes", {})

        assert SPANDATA.CODE_LINE_NUMBER in data
        assert SPANDATA.CODE_NAMESPACE in data
        assert SPANDATA.CODE_FILE_PATH in data
    else:
        events = capture_events()

        with start_transaction(name="test_transaction"):
            client = SyncClientBuilder().build()
            client.get(url).build().send()

        (event,) = events
        span = event["spans"][0]
        data = span.get("data", {})

        assert SPANDATA.CODE_LINENO in data
        assert SPANDATA.CODE_NAMESPACE in data
        assert SPANDATA.CODE_FILEPATH in data
    assert SPANDATA.CODE_FUNCTION in data


@pytest.mark.parametrize("span_streaming", [True, False])
def test_request_source(
    sentry_init,
    capture_events,
    capture_items,
    server_port,
    span_streaming,
):
    sentry_init(
        integrations=[PyreqwestIntegration()],
        traces_sample_rate=1.0,
        enable_http_request_source=True,
        http_request_source_threshold_ms=0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    url = f"http://localhost:{server_port}/hello"
    if span_streaming:
        items = capture_items("span")

        with sentry_sdk.traces.start_span(name="custom parent"):
            client = SyncClientBuilder().build()
            client.get(url).build().send()

        sentry_sdk.flush()
        spans = [item.payload for item in items]
        span = spans[0]
        data = span.get("attributes", {})

        assert type(data.get(SPANDATA.CODE_LINE_NUMBER)) == int
        assert data.get(SPANDATA.CODE_LINE_NUMBER) > 0
        assert (
            data.get(SPANDATA.CODE_NAMESPACE)
            == "tests.integrations.pyreqwest.test_pyreqwest"
        )
        assert data.get(SPANDATA.CODE_FILE_PATH).endswith(
            "tests/integrations/pyreqwest/test_pyreqwest.py"
        )

        is_relative_path = data.get(SPANDATA.CODE_FILE_PATH)[0] != os.sep
    else:
        events = capture_events()

        with start_transaction(name="test_transaction"):
            client = SyncClientBuilder().build()
            client.get(url).build().send()

        (event,) = events
        span = event["spans"][0]
        data = span.get("data", {})

        assert type(data.get(SPANDATA.CODE_LINENO)) == int
        assert data.get(SPANDATA.CODE_LINENO) > 0
        assert (
            data.get(SPANDATA.CODE_NAMESPACE)
            == "tests.integrations.pyreqwest.test_pyreqwest"
        )
        assert data.get(SPANDATA.CODE_FILEPATH).endswith(
            "tests/integrations/pyreqwest/test_pyreqwest.py"
        )

        is_relative_path = data.get(SPANDATA.CODE_FILEPATH)[0] != os.sep
    assert is_relative_path

    assert data.get(SPANDATA.CODE_FUNCTION) == "test_request_source"


@pytest.mark.parametrize("span_streaming", [True, False])
def test_request_source_with_module_in_search_path(
    sentry_init,
    capture_events,
    capture_items,
    server_port,
    span_streaming,
):
    sentry_init(
        integrations=[PyreqwestIntegration()],
        traces_sample_rate=1.0,
        enable_http_request_source=True,
        http_request_source_threshold_ms=0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    url = f"http://localhost:{server_port}/hello"
    if span_streaming:
        items = capture_items("span")

        with sentry_sdk.traces.start_span(name="custom parent"):
            from pyreqwest_helpers.helpers import get_request_with_client

            client = SyncClientBuilder().build()
            get_request_with_client(client, url)

        sentry_sdk.flush()
        spans = [item.payload for item in items]
        span = spans[0]
        data = span.get("attributes", {})

        assert type(data.get(SPANDATA.CODE_LINE_NUMBER)) == int
        assert data.get(SPANDATA.CODE_LINE_NUMBER) > 0
        assert data.get(SPANDATA.CODE_NAMESPACE) == "pyreqwest_helpers.helpers"
        assert data.get(SPANDATA.CODE_FILE_PATH) == "pyreqwest_helpers/helpers.py"

        is_relative_path = data.get(SPANDATA.CODE_FILE_PATH)[0] != os.sep
    else:
        events = capture_events()

        with start_transaction(name="test_transaction"):
            from pyreqwest_helpers.helpers import get_request_with_client

            client = SyncClientBuilder().build()
            get_request_with_client(client, url)

        (event,) = events
        span = event["spans"][0]
        data = span.get("data", {})

        assert type(data.get(SPANDATA.CODE_LINENO)) == int
        assert data.get(SPANDATA.CODE_LINENO) > 0
        assert data.get(SPANDATA.CODE_NAMESPACE) == "pyreqwest_helpers.helpers"
        assert data.get(SPANDATA.CODE_FILEPATH) == "pyreqwest_helpers/helpers.py"

        is_relative_path = data.get(SPANDATA.CODE_FILEPATH)[0] != os.sep
    assert is_relative_path

    assert data.get(SPANDATA.CODE_FUNCTION) == "get_request_with_client"


@pytest.mark.parametrize("span_streaming", [True, False])
def test_no_request_source_if_duration_too_short(
    sentry_init,
    capture_events,
    capture_items,
    server_port,
    span_streaming,
):
    sentry_init(
        integrations=[PyreqwestIntegration()],
        traces_sample_rate=1.0,
        enable_http_request_source=True,
        http_request_source_threshold_ms=100,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    url = f"http://localhost:{server_port}/hello"
    if span_streaming:
        items = capture_items("span")

        with sentry_sdk.traces.start_span(name="custom parent"):
            original_start_span = sentry_sdk.traces.start_span

            @contextmanager
            def fake_start_span(*args, **kwargs):
                with original_start_span(*args, **kwargs) as span:
                    span._start_timestamp = datetime.datetime(2024, 1, 1, microsecond=0)
                    span._end_timestamp = datetime.datetime(
                        2024, 1, 1, microsecond=99999
                    )

                    yield span

                    span._end_timestamp = None

            with mock.patch(
                "sentry_sdk.integrations.pyreqwest.sentry_sdk.traces.start_span",
                fake_start_span,
            ):
                client = SyncClientBuilder().build()
                client.get(url).build().send()

        sentry_sdk.flush()
        spans = [item.payload for item in items]
        span = spans[0]
        data = span.get("attributes", {})

        assert SPANDATA.CODE_LINE_NUMBER not in data
        assert SPANDATA.CODE_NAMESPACE not in data
        assert SPANDATA.CODE_FILE_PATH not in data
    else:
        events = capture_events()

        with start_transaction(name="test_transaction"):

            @contextmanager
            def fake_start_span(*args, **kwargs):
                with sentry_sdk.start_span(*args, **kwargs) as span:
                    pass
                span.start_timestamp = datetime.datetime(2024, 1, 1, microsecond=0)
                span.timestamp = datetime.datetime(2024, 1, 1, microsecond=99999)
                yield span

            with mock.patch(
                "sentry_sdk.integrations.pyreqwest.start_span",
                fake_start_span,
            ):
                client = SyncClientBuilder().build()
                client.get(url).build().send()

        (event,) = events
        span = event["spans"][-1]
        data = span.get("data", {})

        assert SPANDATA.CODE_LINENO not in data
        assert SPANDATA.CODE_NAMESPACE not in data
        assert SPANDATA.CODE_FILEPATH not in data
    assert SPANDATA.CODE_FUNCTION not in data


@pytest.mark.parametrize("span_streaming", [True, False])
def test_request_source_if_duration_over_threshold(
    sentry_init,
    capture_events,
    capture_items,
    server_port,
    span_streaming,
):
    sentry_init(
        integrations=[PyreqwestIntegration()],
        traces_sample_rate=1.0,
        enable_http_request_source=True,
        http_request_source_threshold_ms=100,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    url = f"http://localhost:{server_port}/hello"
    if span_streaming:
        items = capture_items("span")

        with sentry_sdk.traces.start_span(name="custom parent"):
            original_start_span = sentry_sdk.traces.start_span

            @contextmanager
            def fake_start_span(*args, **kwargs):
                with original_start_span(*args, **kwargs) as span:
                    span._start_timestamp = datetime.datetime(2024, 1, 1, microsecond=0)
                    span._end_timestamp = datetime.datetime(
                        2024, 1, 1, microsecond=100001
                    )

                    yield span

                    span._end_timestamp = None

            with mock.patch(
                "sentry_sdk.integrations.pyreqwest.sentry_sdk.traces.start_span",
                fake_start_span,
            ):
                client = SyncClientBuilder().build()
                client.get(url).build().send()

        sentry_sdk.flush()
        spans = [item.payload for item in items]
        span = spans[0]
        data = span.get("attributes", {})

        assert SPANDATA.CODE_LINE_NUMBER in data
        assert SPANDATA.CODE_NAMESPACE in data
        assert SPANDATA.CODE_FILE_PATH in data
    else:
        events = capture_events()

        with start_transaction(name="test_transaction"):

            @contextmanager
            def fake_start_span(*args, **kwargs):
                with sentry_sdk.start_span(*args, **kwargs) as span:
                    pass
                span.start_timestamp = datetime.datetime(2024, 1, 1, microsecond=0)
                span.timestamp = datetime.datetime(2024, 1, 1, microsecond=100001)
                yield span

            with mock.patch(
                "sentry_sdk.integrations.pyreqwest.start_span",
                fake_start_span,
            ):
                client = SyncClientBuilder().build()
                client.get(url).build().send()

        (event,) = events
        span = event["spans"][-1]
        data = span.get("data", {})

        assert SPANDATA.CODE_LINENO in data
        assert SPANDATA.CODE_NAMESPACE in data
        assert SPANDATA.CODE_FILEPATH in data
    assert SPANDATA.CODE_FUNCTION in data



import pytest

from sentry_sdk.integrations.serverless import serverless_function


def test_basic(sentry_init, capture_exceptions, monkeypatch):
    sentry_init()
    exceptions = capture_exceptions()

    flush_calls = []

    @serverless_function
    def foo():
        monkeypatch.setattr("sentry_sdk.flush", lambda: flush_calls.append(1))
        1 / 0

    with pytest.raises(ZeroDivisionError):
        foo()

    (exception,) = exceptions
    assert isinstance(exception, ZeroDivisionError)

    assert flush_calls == [1]


def test_flush_disabled(sentry_init, capture_exceptions, monkeypatch):
    sentry_init()
    exceptions = capture_exceptions()

    flush_calls = []

    monkeypatch.setattr("sentry_sdk.flush", lambda: flush_calls.append(1))

    @serverless_function(flush=False)
    def foo():
        1 / 0

    with pytest.raises(ZeroDivisionError):
        foo()

    (exception,) = exceptions
    assert isinstance(exception, ZeroDivisionError)

    assert flush_calls == []



from functools import wraps
from typing import (
    Any,
    AsyncIterator,
    Callable,
    Iterator,
    List,
)

import sentry_sdk
from sentry_sdk.ai.utils import get_start_span_function
from sentry_sdk.consts import OP, SPANDATA
from sentry_sdk.integrations import DidNotEnable, Integration
from sentry_sdk.traces import SpanStatus, StreamedSpan
from sentry_sdk.tracing import SPANSTATUS
from sentry_sdk.tracing_utils import has_span_streaming_enabled

try:
    from google.genai.models import AsyncModels, Models
except ImportError:
    raise DidNotEnable("google-genai not installed")


from .consts import GEN_AI_SYSTEM, IDENTIFIER, ORIGIN
from .streaming import (
    accumulate_streaming_response,
    set_span_data_for_streaming_response,
)
from .utils import (
    _capture_exception,
    prepare_embed_content_args,
    prepare_generate_content_args,
    set_span_data_for_embed_request,
    set_span_data_for_embed_response,
    set_span_data_for_request,
    set_span_data_for_response,
)


class GoogleGenAIIntegration(Integration):
    identifier = IDENTIFIER
    origin = ORIGIN

    def __init__(self: "GoogleGenAIIntegration", include_prompts: bool = True) -> None:
        self.include_prompts = include_prompts

    @staticmethod
    def setup_once() -> None:
        # Patch sync methods
        Models.generate_content = _wrap_generate_content(Models.generate_content)
        Models.generate_content_stream = _wrap_generate_content_stream(
            Models.generate_content_stream
        )
        Models.embed_content = _wrap_embed_content(Models.embed_content)

        # Patch async methods
        AsyncModels.generate_content = _wrap_async_generate_content(
            AsyncModels.generate_content
        )
        AsyncModels.generate_content_stream = _wrap_async_generate_content_stream(
            AsyncModels.generate_content_stream
        )
        AsyncModels.embed_content = _wrap_async_embed_content(AsyncModels.embed_content)


def _wrap_generate_content_stream(f: "Callable[..., Any]") -> "Callable[..., Any]":
    @wraps(f)
    def new_generate_content_stream(
        self: "Any", *args: "Any", **kwargs: "Any"
    ) -> "Any":
        client = sentry_sdk.get_client()
        integration = client.get_integration(GoogleGenAIIntegration)
        if integration is None:
            return f(self, *args, **kwargs)

        _model, contents, model_name = prepare_generate_content_args(args, kwargs)

        if has_span_streaming_enabled(client.options):
            chat_span = sentry_sdk.traces.start_span(
                name=f"chat {model_name}",
                attributes={
                    "sentry.op": OP.GEN_AI_CHAT,
                    "sentry.origin": ORIGIN,
                    SPANDATA.GEN_AI_OPERATION_NAME: "chat",
                    SPANDATA.GEN_AI_SYSTEM: GEN_AI_SYSTEM,
                    SPANDATA.GEN_AI_REQUEST_MODEL: model_name,
                    SPANDATA.GEN_AI_RESPONSE_STREAMING: True,
                },
            )
        else:
            chat_span = get_start_span_function()(
                op=OP.GEN_AI_CHAT,
                name=f"chat {model_name}",
                origin=ORIGIN,
            )
            chat_span.__enter__()

            chat_span.set_data(SPANDATA.GEN_AI_OPERATION_NAME, "chat")
            chat_span.set_data(SPANDATA.GEN_AI_SYSTEM, GEN_AI_SYSTEM)
            chat_span.set_data(SPANDATA.GEN_AI_REQUEST_MODEL, model_name)
            chat_span.set_data(SPANDATA.GEN_AI_RESPONSE_STREAMING, True)

        set_span_data_for_request(chat_span, integration, model_name, contents, kwargs)

        try:
            stream = f(self, *args, **kwargs)

            # Create wrapper iterator to accumulate responses
            def new_iterator() -> "Iterator[Any]":
                chunks: "List[Any]" = []
                try:
                    for chunk in stream:
                        chunks.append(chunk)
                        yield chunk
                except Exception as exc:
                    _capture_exception(exc)
                    if isinstance(chat_span, StreamedSpan):
                        chat_span.status = SpanStatus.ERROR
                    else:
                        chat_span.set_status(SPANSTATUS.INTERNAL_ERROR)
                    raise
                finally:
                    # Accumulate all chunks and set final response data on spans
                    if chunks:
                        accumulated_response = accumulate_streaming_response(chunks)
                        set_span_data_for_streaming_response(
                            chat_span, integration, accumulated_response
                        )
                    chat_span.__exit__(None, None, None)

            return new_iterator()

        except Exception as exc:
            _capture_exception(exc)
            chat_span.__exit__(None, None, None)
            raise

    return new_generate_content_stream


def _wrap_async_generate_content_stream(
    f: "Callable[..., Any]",
) -> "Callable[..., Any]":
    @wraps(f)
    async def new_async_generate_content_stream(
        self: "Any", *args: "Any", **kwargs: "Any"
    ) -> "Any":
        client = sentry_sdk.get_client()
        integration = client.get_integration(GoogleGenAIIntegration)
        if integration is None:
            return await f(self, *args, **kwargs)

        _model, contents, model_name = prepare_generate_content_args(args, kwargs)

        if has_span_streaming_enabled(client.options):
            chat_span = sentry_sdk.traces.start_span(
                name=f"chat {model_name}",
                attributes={
                    "sentry.op": OP.GEN_AI_CHAT,
                    "sentry.origin": ORIGIN,
                    SPANDATA.GEN_AI_OPERATION_NAME: "chat",
                    SPANDATA.GEN_AI_SYSTEM: GEN_AI_SYSTEM,
                    SPANDATA.GEN_AI_REQUEST_MODEL: model_name,
                    SPANDATA.GEN_AI_RESPONSE_STREAMING: True,
                },
            )
        else:
            chat_span = get_start_span_function()(
                op=OP.GEN_AI_CHAT,
                name=f"chat {model_name}",
                origin=ORIGIN,
            )
            chat_span.__enter__()

            chat_span.set_data(SPANDATA.GEN_AI_OPERATION_NAME, "chat")
            chat_span.set_data(SPANDATA.GEN_AI_SYSTEM, GEN_AI_SYSTEM)
            chat_span.set_data(SPANDATA.GEN_AI_REQUEST_MODEL, model_name)
            chat_span.set_data(SPANDATA.GEN_AI_RESPONSE_STREAMING, True)

        set_span_data_for_request(chat_span, integration, model_name, contents, kwargs)

        try:
            stream = await f(self, *args, **kwargs)

            # Create wrapper async iterator to accumulate responses
            async def new_async_iterator() -> "AsyncIterator[Any]":
                chunks: "List[Any]" = []
                try:
                    async for chunk in stream:
                        chunks.append(chunk)
                        yield chunk
                except Exception as exc:
                    _capture_exception(exc)
                    if isinstance(chat_span, StreamedSpan):
                        chat_span.status = SpanStatus.ERROR
                    else:
                        chat_span.set_status(SPANSTATUS.INTERNAL_ERROR)
                    raise
                finally:
                    # Accumulate all chunks and set final response data on spans
                    if chunks:
                        accumulated_response = accumulate_streaming_response(chunks)
                        set_span_data_for_streaming_response(
                            chat_span, integration, accumulated_response
                        )
                    chat_span.__exit__(None, None, None)

            return new_async_iterator()

        except Exception as exc:
            _capture_exception(exc)
            chat_span.__exit__(None, None, None)
            raise

    return new_async_generate_content_stream


def _wrap_generate_content(f: "Callable[..., Any]") -> "Callable[..., Any]":
    @wraps(f)
    def new_generate_content(self: "Any", *args: "Any", **kwargs: "Any") -> "Any":
        client = sentry_sdk.get_client()
        integration = client.get_integration(GoogleGenAIIntegration)
        if integration is None:
            return f(self, *args, **kwargs)

        model, contents, model_name = prepare_generate_content_args(args, kwargs)

        if has_span_streaming_enabled(client.options):
            with sentry_sdk.traces.start_span(
                name=f"chat {model_name}",
                attributes={
                    "sentry.op": OP.GEN_AI_CHAT,
                    "sentry.origin": ORIGIN,
                    SPANDATA.GEN_AI_OPERATION_NAME: "chat",
                    SPANDATA.GEN_AI_SYSTEM: GEN_AI_SYSTEM,
                    SPANDATA.GEN_AI_REQUEST_MODEL: model_name,
                },
            ) as chat_span:
                set_span_data_for_request(
                    chat_span, integration, model_name, contents, kwargs
                )

                try:
                    response = f(self, *args, **kwargs)
                except Exception as exc:
                    _capture_exception(exc)
                    chat_span.status = SpanStatus.ERROR
                    raise

                set_span_data_for_response(chat_span, integration, response)

                return response
        else:
            with get_start_span_function()(
                op=OP.GEN_AI_CHAT,
                name=f"chat {model_name}",
                origin=ORIGIN,
            ) as chat_span:
                chat_span.set_data(SPANDATA.GEN_AI_OPERATION_NAME, "chat")
                chat_span.set_data(SPANDATA.GEN_AI_SYSTEM, GEN_AI_SYSTEM)
                chat_span.set_data(SPANDATA.GEN_AI_REQUEST_MODEL, model_name)
                set_span_data_for_request(
                    chat_span, integration, model_name, contents, kwargs
                )

                try:
                    response = f(self, *args, **kwargs)
                except Exception as exc:
                    _capture_exception(exc)
                    chat_span.set_status(SPANSTATUS.INTERNAL_ERROR)
                    raise

                set_span_data_for_response(chat_span, integration, response)

                return response

    return new_generate_content


def _wrap_async_generate_content(f: "Callable[..., Any]") -> "Callable[..., Any]":
    @wraps(f)
    async def new_async_generate_content(
        self: "Any", *args: "Any", **kwargs: "Any"
    ) -> "Any":
        client = sentry_sdk.get_client()
        integration = client.get_integration(GoogleGenAIIntegration)
        if integration is None:
            return await f(self, *args, **kwargs)

        model, contents, model_name = prepare_generate_content_args(args, kwargs)

        if has_span_streaming_enabled(client.options):
            with sentry_sdk.traces.start_span(
                name=f"chat {model_name}",
                attributes={
                    "sentry.op": OP.GEN_AI_CHAT,
                    "sentry.origin": ORIGIN,
                    SPANDATA.GEN_AI_OPERATION_NAME: "chat",
                    SPANDATA.GEN_AI_SYSTEM: GEN_AI_SYSTEM,
                    SPANDATA.GEN_AI_REQUEST_MODEL: model_name,
                },
            ) as chat_span:
                set_span_data_for_request(
                    chat_span, integration, model_name, contents, kwargs
                )
                try:
                    response = await f(self, *args, **kwargs)
                except Exception as exc:
                    _capture_exception(exc)
                    chat_span.status = SpanStatus.ERROR
                    raise

                set_span_data_for_response(chat_span, integration, response)

                return response
        else:
            with get_start_span_function()(
                op=OP.GEN_AI_CHAT,
                name=f"chat {model_name}",
                origin=ORIGIN,
            ) as chat_span:
                chat_span.set_data(SPANDATA.GEN_AI_OPERATION_NAME, "chat")
                chat_span.set_data(SPANDATA.GEN_AI_SYSTEM, GEN_AI_SYSTEM)
                chat_span.set_data(SPANDATA.GEN_AI_REQUEST_MODEL, model_name)
                set_span_data_for_request(
                    chat_span, integration, model_name, contents, kwargs
                )
                try:
                    response = await f(self, *args, **kwargs)
                except Exception as exc:
                    _capture_exception(exc)
                    chat_span.set_status(SPANSTATUS.INTERNAL_ERROR)
                    raise

                set_span_data_for_response(chat_span, integration, response)

                return response

    return new_async_generate_content


def _wrap_embed_content(f: "Callable[..., Any]") -> "Callable[..., Any]":
    @wraps(f)
    def new_embed_content(self: "Any", *args: "Any", **kwargs: "Any") -> "Any":
        client = sentry_sdk.get_client()
        integration = client.get_integration(GoogleGenAIIntegration)
        if integration is None:
            return f(self, *args, **kwargs)

        model_name, contents = prepare_embed_content_args(args, kwargs)

        if has_span_streaming_enabled(client.options):
            with sentry_sdk.traces.start_span(
                name=f"embeddings {model_name}",
                attributes={
                    "sentry.op": OP.GEN_AI_EMBEDDINGS,
                    "sentry.origin": ORIGIN,
                    SPANDATA.GEN_AI_OPERATION_NAME: "embeddings",
                    SPANDATA.GEN_AI_SYSTEM: GEN_AI_SYSTEM,
                    SPANDATA.GEN_AI_REQUEST_MODEL: model_name,
                },
            ) as span:
                set_span_data_for_embed_request(span, integration, contents, kwargs)

                try:
                    response = f(self, *args, **kwargs)
                except Exception as exc:
                    _capture_exception(exc)
                    span.status = SpanStatus.ERROR
                    raise

                set_span_data_for_embed_response(span, integration, response)

                return response
        else:
            with get_start_span_function()(
                op=OP.GEN_AI_EMBEDDINGS,
                name=f"embeddings {model_name}",
                origin=ORIGIN,
            ) as span:
                span.set_data(SPANDATA.GEN_AI_OPERATION_NAME, "embeddings")
                span.set_data(SPANDATA.GEN_AI_SYSTEM, GEN_AI_SYSTEM)
                span.set_data(SPANDATA.GEN_AI_REQUEST_MODEL, model_name)
                set_span_data_for_embed_request(span, integration, contents, kwargs)

                try:
                    response = f(self, *args, **kwargs)
                except Exception as exc:
                    _capture_exception(exc)
                    span.set_status(SPANSTATUS.INTERNAL_ERROR)
                    raise

                set_span_data_for_embed_response(span, integration, response)

                return response

    return new_embed_content


def _wrap_async_embed_content(f: "Callable[..., Any]") -> "Callable[..., Any]":
    @wraps(f)
    async def new_async_embed_content(
        self: "Any", *args: "Any", **kwargs: "Any"
    ) -> "Any":
        client = sentry_sdk.get_client()
        integration = client.get_integration(GoogleGenAIIntegration)
        if integration is None:
            return await f(self, *args, **kwargs)

        model_name, contents = prepare_embed_content_args(args, kwargs)

        if has_span_streaming_enabled(client.options):
            with sentry_sdk.traces.start_span(
                name=f"embeddings {model_name}",
                attributes={
                    "sentry.op": OP.GEN_AI_EMBEDDINGS,
                    "sentry.origin": ORIGIN,
                    SPANDATA.GEN_AI_OPERATION_NAME: "embeddings",
                    SPANDATA.GEN_AI_SYSTEM: GEN_AI_SYSTEM,
                    SPANDATA.GEN_AI_REQUEST_MODEL: model_name,
                },
            ) as span:
                set_span_data_for_embed_request(span, integration, contents, kwargs)

                try:
                    response = await f(self, *args, **kwargs)
                except Exception as exc:
                    _capture_exception(exc)
                    span.status = SpanStatus.ERROR
                    raise

                set_span_data_for_embed_response(span, integration, response)

                return response
        else:
            with get_start_span_function()(
                op=OP.GEN_AI_EMBEDDINGS,
                name=f"embeddings {model_name}",
                origin=ORIGIN,
            ) as span:
                span.set_data(SPANDATA.GEN_AI_OPERATION_NAME, "embeddings")
                span.set_data(SPANDATA.GEN_AI_SYSTEM, GEN_AI_SYSTEM)
                span.set_data(SPANDATA.GEN_AI_REQUEST_MODEL, model_name)
                set_span_data_for_embed_request(span, integration, contents, kwargs)

                try:
                    response = await f(self, *args, **kwargs)
                except Exception as exc:
                    _capture_exception(exc)
                    span.set_status(SPANSTATUS.INTERNAL_ERROR)
                    raise

                set_span_data_for_embed_response(span, integration, response)

                return response

    return new_async_embed_content
